/*****************************************************************************
 * $Id: game_pony_sky.cc,v 1.30 2006/08/10 03:18:59 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	Sky class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#define GAME_PONY_SKY_CC 1
#include "game_pony_common.hh"
#include "game_pony_sky.hh"
#include "game_pony_cloud.hh"
#include "gfx_random.hh"

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////  Sky  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Sky constructor.
 *****************************************************************************/
Sky::Sky( QuadrantPtrc q )
	: Omni(q), mPrevEye(0,0,0)
{
	if ( ! gWorldParm.mTerrain.mDesert )
	{
		// Randomly make clouds in this quadrant.
		const uint cloudCnt = int(gWorldParm.mObjRatio * (70 + random_ui(30)));
		for ( uint i = 0; i < cloudCnt; ++i )
		{
			Position pos;
			q->RandomPosition( pos );
			pos.y = WORLD(3000.0 + random_f(1000.0));
			new PerlinCloud( q, pos );
		}
	}
	DEBUG_CODE( CheckObjectCtor(); )
}
Sky::~Sky()
{
	DEBUG_CODE( CheckObjectDtor(); )
}

/*****************************************************************************
 * Draw sky.
 *****************************************************************************/
void
Sky::Draw( void )
{
	DrawSkyCylinder();
}

/*****************************************************************************
 * Draw sky as a cylinder but with the top circle reduced (a frustrum/cylider).
 *****************************************************************************/
void
Sky::DrawSkyCylinder( void )
{
	// The sky cylinder encloses the tic-tac-toe quadrants.
	// radius = quadrantEdge * 2.0 (approx)
	// +------+------+------+------+------+
	// |     +|------|------|------|+     |
	// |    / |      |      |      | \    |
	// +------+======+======+======+------+
	// |   |  ||     |      |     ||  |   |
	// |   |  ||     |      |     ||  |   |
	// +------+------+------+------+------+
	// |   |  ||     |  .   |     ||  |   |
	// |   |  ||     | eye  |     ||  |   |
	// +------+------+------+------+------+
	// |   |  ||     |      |     ||  |   |
	// |   |  ||     |      |     ||  |   |
	// +------+======+======+======+------+
	// |    \ |      |      |      | /    |
	// |     +|------|------|------|+     |
	// +------+------+------+------+------+
#if 0
	fp					radius		= gWorld.GetQuadrantEdge() * 2.0;
	const fp&			height		= gWorld.GetQuadrantEdge();
#else
	// As a workaround for Voodoo3 on X/Mesa, reduce the size of the sky box,
	// Otherwise, fog color won't be right (on Voodoo 3).
	// See doc/issues.txt for more.
	fp					radius		= 0.40 * gWorld.GetQuadrantEdge();
	const fp&			height		= 0.20 * gWorld.GetQuadrantEdge();
#endif
	static const RGBA	rgbaBlue( 0.0, 0.0, 1.0, 1.0 );
	static const RGBA	rgbaSky( RGB_SKYBLUE, 1.0 );
	int					angle, i;

	// Calculate vertexes of cylinder (in its own coord system).
	if ( gEng.mEye.IsRecalcNeeded( mPrevEye, 50 ) )
	{
		for ( angle=0,i=0; angle <= 360; angle += 360/mSlices,++i )
		{
			// cylinder bottom
			mVg[i].x = cos_deg(angle) * radius - gEng.mEye.mCm[Ox];
			mVg[i].y = ALTITUDE_LAND;
			mVg[i].z = sin_deg(angle) * radius - gEng.mEye.mCm[Oz];

			// cylinder top
			mVs[i].x = cos_deg(angle) * radius - gEng.mEye.mCm[Ox];
			mVs[i].y = height;
			mVs[i].z = sin_deg(angle) * radius - gEng.mEye.mCm[Oz];
		}
	}

	bool savedDepthMask = GFX_DEPTH_MASK( false );

	GFX_RGBA( gEng.GetFogDensity() >= 1.0 ? rgbaBlue : rgbaSky );

	// Transform/submit sides of cylinder as quad strips.
   	GFX_BEGIN( GP_QUAD_STRIP );
	for ( angle=0,i=0; angle <= 360; angle += 360/mSlices,++i )
	{
		GFX_TRANSFORM_VERTEX( mVg[i] );
		GFX_TRANSFORM_VERTEX( mVs[i] );
	}
	GFX_END();

	// Transform/submit top of cylinder (circle).
   	GFX_BEGIN( GP_POLYGON );
	for ( angle=0,i=0; angle < 360; angle += 360/mSlices,i++ )
	{
		GFX_TRANSFORM_VERTEX( mVs[i] );
	}
	GFX_END();

	GFX_DEPTH_MASK( savedDepthMask );
}

/*****************************************************************************
 * @return distance^2
 *****************************************************************************/
fp
Sky::Distance2( void )
{
	// See Terrain::Distance2().
	return Parent::Distance2();
}
